/*
 * Denali Nand Driver.
 *
 * Copyright (C) 2010 Renesas Electronics Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/device.h>
#include <linux/io.h>
#include <linux/platform_device.h>
#include <linux/leds.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/nand_platform.h>

/*
 * Registers offset
 */

/* Nand Flash Memory Controller Registers */
#define NANDF_DEVICE_RESET				0x0
#define NANDF_TRANSFER_SPARE				0x10
#define NANDF_LOAD_WAIT_CNT				0x20
#define NANDF_PROGRAM_WAIT_CNT				0x30
#define NANDF_ERASE_WAIT_CNT				0x40
#define NANDF_INT_MON_CYCCNT				0x50
#define NANDF_RD_PIN_EN					0x60
#define NANDF_MULTIPLANE_OPE				0x70
#define NANDF_MULTIPLANE_RD_EN				0x80
#define NANDF_COPYBACK_DISABLE				0x90
#define NANDF_CACHE_WR_EN				0xa0
#define NANDF_CACHE_RD_EN				0xb0
#define NANDF_PREFETCH_MODE				0xc0
#define NANDF_CHIP_EN_DONT_CARE				0xd0
#define NANDF_ECC_EN					0xe0
#define NANDF_GLOBAL_INT_EN				0xf0
#define NANDF_WE_2_RE					0x100
#define NANDF_ADDR_2_DATA				0x110
#define NANDF_RE_2_WE					0x120
#define NANDF_ACC_CLKS					0x130
#define NANDF_NUM_OF_PLANES				0x140
#define NANDF_PAGES_PER_BLOCK				0x150
#define NANDF_DEVICE_WIDTH				0x160
#define NANDF_DEV_MAIN_AREA_SIZE			0x170
#define NANDF_DEV_SPARE_AREA_SIZE			0x180
#define NANDF_TWO_ROW_ADDR_CYCLE			0x190
#define NANDF_MULTI_ADDR_REST				0x1a0
#define NANDF_ECC_CORRECT				0x1b0
#define NANDF_READ_MODE					0x1c0
#define NANDF_WRITE_MODE				0x1d0
#define NANDF_COPYBACK_MODE				0x1e0
#define NANDF_RDWR_EN_LO_CNT				0x1f0
#define NANDF_RDWR_EN_HI_CNT				0x200
#define NANDF_MAX_RD_DELAY				0x210
#define NANDF_CS_SETUP_CNT				0x220
#define NANDF_SPARE_AREA_SKIP_BYTE			0x230
#define NANDF_SPARE_AREA_MARKER				0x240
#define NANDF_DEVICES_CONNECT				0x250
#define NANDF_DIE_MASK					0x260
#define NANDF_1ST_BLOCK_OF_NEXT_PLANE			0x270
#define NANDF_WRITE_PROTECT				0x280
#define NANDF_RE_2_RE					0x290
#define NANDF_POR_RESET_CNT				0x2a0
#define NANDF_WATCHDOC_RESET_CNT			0x2b0
#define NANDF_MANUFACTURE_ID				0x300
#define NANDF_DEVICE_ID					0x310
#define NANDF_DEVICE_PARAM_0				0x320
#define NANDF_DEVICE_PARAM_1				0x330
#define NANDF_DEVICE_PARAM_2				0x340
#define NANDF_LOGICAL_PAGE_DATA_SIZE			0x350
#define NANDF_LOGICAL_PAGE_SPARE_SIZE			0x360
#define NANDF_REVISION					0x370
#define NANDF_ONFI_DEVICE_FEATURES			0x380
#define NANDF_ONFI_OPTIONAL_COMMAND			0x390
#define NANDF_ONFI_TIMING_MODE				0x3a0
#define NANDF_ONFI_PGM_CACHE_TIMING_MODE		0x3b0
#define NANDF_ONFI_DEVICE_NO_OF_LUNS			0x3c0
#define NANDF_ONFI_DEVICE_NO_OF_BLOKS_PER_LUN_L		0x3d0
#define NANDF_ONFI_DEVICE_NO_OF_BLOCKS_PER_LUN_U	0x3e0
#define NANDF_FEATURES					0x3f0
#define NANDF_TRANSFER_MODE				0x400
#define NANDF_INTR_STATUS0				0x410
#define NANDF_INTR_EN0					0x420
#define NANDF_PAGE_CNT0					0x430
#define NANDF_ERR_PAGE_ADDR0				0x440
#define NANDF_ERR_BROCK_ADDR0				0x450
#define NANDF_INTR_STATUS1				0x460
#define NANDF_INTR_EN1					0x470
#define NANDF_PAGE_CNT1					0x480
#define NANDF_ERR_PAGE_ADDR1				0x490
#define NANDF_ERR_BLOCK_ADDR1				0x4a0
#define NANDF_ECCCORINFO_B01				0x650
#define NANDF_DMA_ENABLE				0x700
#define NANDF_DMA_INTR					0x720
#define NANDF_DMA_INTR_EN				0x730
#define NANDF_TARGET_ERR_ADDR_LO			0x740
#define NANDF_TARGET_ERR_ADDR_HI			0x750
#define NANDF_CHNL_ACTIVE				0x760
#define NANDF_FLASH_BURST_LENGTH			0x770
#define NANDF_CHIP_INTERLEAVE_ENABLE			0x780
#define NANDF_RESCAN_BUFFER_FLAG			0x790

/* Nand Flash Memory Controller Indexed-Addressing Registers */
#define NANDF_INDEX_CONTROL				0x0
#define NANDF_INDEX_DATA				0x10

/* Nand Flash Local Bridge Registers */
#define NANDF_AHB_ERROR					0x8
#define NANDF_RESET					0x18
#define NANDF_AHB_STATUS				0x1c
#define NANDF_BOOTSTRAP_INFO				0x28
#define NANDF_SRAM_CTR					0x30
#define NANDF_SRAM_WADR					0x34

/* Command Mapping */
#define MAP00						0x0
#define MAP01						0x1
#define MAP10						0x2
#define MAP11						0x3

/* CMD CYCLE */
#define COMMAND_CYCLE					0x0
#define ADDRESS_CYCLE					0x1
#define STATUS_CYCLE					0x2

#define DMA_COMP					0x4
#define ERASE_COMP					0x100
#define INT_ACT						0x1000
#define READ						0x0
#define WRITE						0x1
#define SECTOR_SIZE					512

/*
 * MTD structure
 */
struct denali_nand_prv {
	struct mtd_info		mtd;
	struct nand_chip	chip;
	int			irq;
	int			ecc;
	void __iomem		*c_regs;
	void __iomem		*i_regs;
	void __iomem		*b_regs;
	int			chipsel;
	struct completion	cmd_complete;
};

/*
 * HW Control
 */
struct nand_hw_control denali_hwcontrol;

/*
 * Page Buffers
 */
uint8_t buffers[4096 + 128] __attribute__((aligned(32)));

/*
 * Debug
 */
#define NAND_DEBUG			0

#if NAND_DEBUG

static int nand_debug;

#define DPRINTK(level, args...)				\
		do {					\
			if ((level) <= nand_debug)	\
				printk(KERN_INFO args);	\
		} while (0)
#else
#define DPRINTK(level, args...)
#endif				/* NAND_DEBUG */

static struct nand_ecclayout nand_oob_64 = {
	.eccbytes = 32,
	.oobfree = {
		{.offset = 32,
		.length = 32} }
};

static struct nand_ecclayout nand_oob_128 = {
	.eccbytes = 64,
	.oobfree = {
		{.offset = 64,
		.length = 64} }
};

#ifdef CONFIG_MTD_PARTITIONS
#ifdef CONFIG_MTD_CMDLINE_PARTS
static const char *partition_probes[] = { "cmdlinepart", NULL };
#endif
#endif

/* flash bbt decriptors for large page devices */
static uint8_t bbt_pattern[] = {'B', 'b', 't', '0' };
static uint8_t mirror_pattern[] = {'1', 't', 'b', 'B' };

static struct nand_bbt_descr bbt_main_descr = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_CREATE | NAND_BBT_WRITE |
		  NAND_BBT_2BIT | NAND_BBT_VERSION | NAND_BBT_PERCHIP,
	.offs =	56,
	.len = 4,
	.veroffs = 60,
	.maxblocks = 4,
	.pattern = bbt_pattern
};

static struct nand_bbt_descr bbt_mirror_descr = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_CREATE | NAND_BBT_WRITE |
		  NAND_BBT_2BIT | NAND_BBT_VERSION | NAND_BBT_PERCHIP,
	.offs =	56,
	.len = 4,
	.veroffs = 60,
	.maxblocks = 4,
	.pattern = mirror_pattern
};

static struct nand_bbt_descr bbt_main_descr_vlp = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_CREATE | NAND_BBT_WRITE |
		  NAND_BBT_2BIT | NAND_BBT_VERSION | NAND_BBT_PERCHIP,
	.offs =	112,
	.len = 4,
	.veroffs = 116,
	.maxblocks = 4,
	.pattern = bbt_pattern
};

static struct nand_bbt_descr bbt_mirror_descr_vlp = {
	.options = NAND_BBT_LASTBLOCK | NAND_BBT_CREATE | NAND_BBT_WRITE |
		  NAND_BBT_2BIT | NAND_BBT_VERSION | NAND_BBT_PERCHIP,
	.offs =	112,
	.len = 4,
	.veroffs = 116,
	.maxblocks = 4,
	.pattern = mirror_pattern
};

/* Read Controller register */
static inline u32 denali_readl(struct nand_chip *chip, u32 reg)
{
	struct denali_nand_prv *prv = chip->priv;

	return readl(prv->c_regs + reg);
}

/* Write Controller register */
static inline void denali_writel(struct nand_chip *chip, u32 reg, u32 val)
{
	struct denali_nand_prv *prv = chip->priv;

	writel(val, prv->c_regs + reg);
}

/* Write Index Control register */
static inline void denali_index_control_set(struct nand_chip *chip, u32 val)
{
	struct denali_nand_prv *prv = chip->priv;

	writel(val, prv->i_regs + NANDF_INDEX_CONTROL);
}

/* Read Index data register */
static inline u32 denali_index_data_read(struct nand_chip *chip)
{
	struct denali_nand_prv *prv = chip->priv;

	return readl(prv->i_regs + NANDF_INDEX_DATA);
}

/* Write Index data register */
static inline void denali_index_data_write(struct nand_chip *chip, u32 val)
{
	struct denali_nand_prv *prv = chip->priv;

	writel(val, prv->i_regs + NANDF_INDEX_DATA);
}

static void denali_nand_select_chip(struct mtd_info *mtd, int chip)
{
	struct nand_chip *chips = mtd->priv;

	if (mtd->writesize != 0) {
		/* Number of pages in a block */
		denali_writel(chips, NANDF_PAGES_PER_BLOCK,
			mtd->erasesize / mtd->writesize);

		/* Set Page main area size */
		denali_writel(chips, NANDF_DEV_MAIN_AREA_SIZE,
			mtd->writesize);

		/* Set Page spare area size */
		denali_writel(chips, NANDF_DEV_SPARE_AREA_SIZE,
			mtd->oobsize);
	}
}

static uint8_t denali_nand_read_byte(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;

	return (uint8_t)denali_index_data_read(chip);
}

static u16 denali_nand_read_word(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;

	return (u16)denali_index_data_read(chip);
}

static void denali_nand_write_buf(struct mtd_info *mtd,
				const uint8_t *buf,
				int len)
{
	int i;
	struct nand_chip *chip = mtd->priv;

	for (i = 0; i < len; i++)
		denali_index_data_write(chip, buf[i]);
}

static void denali_nand_read_buf(struct mtd_info *mtd, uint8_t *buf, int len)
{
	int i;
	struct nand_chip *chip = mtd->priv;

	for (i = 0; i < len; i++)
		buf[i] = (uint8_t)denali_index_data_read(chip);
}

static void denali_nand_cmd_ctrl(struct mtd_info *mtd,
				  int dat,
				  unsigned int ctrl)
{
}

static int denali_nand_dev_ready(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;
	struct denali_nand_prv *prv = chip->priv;

	/* R/B pin 1:READY 0:Busy */
	if (prv->chipsel == 0) {
		/* BANK0 */
		while (!(denali_readl(chip, NANDF_INTR_STATUS0) & INT_ACT))
			;
		/* Interrupt status register clear */
		denali_writel(chip, NANDF_INTR_STATUS0, 0x1000);
	} else {
		/* BANK1 */
		while (!(denali_readl(chip, NANDF_INTR_STATUS1) & INT_ACT))
			;
		/* Interrupt status register clear */
		denali_writel(chip, NANDF_INTR_STATUS1, 0x1000);
	}
	return 0;
}

static void wait_erase_complete(struct nand_chip *chip)
{
	struct denali_nand_prv *prv = chip->priv;

	if (prv->chipsel == 0) {
		/* BANK0 */
		while (!(denali_readl(chip, NANDF_INTR_STATUS0) & ERASE_COMP))
			;
		/* Interrupt status register clear */
		denali_writel(chip, NANDF_INTR_STATUS0, 0x100);
	} else {
		/* BANK1 */
		while (!(denali_readl(chip, NANDF_INTR_STATUS1) & ERASE_COMP))
			;
		/* Interrupt status register clear */
		denali_writel(chip, NANDF_INTR_STATUS1, 0x100);
	}
}

static int nand_dma_transfer(struct nand_chip *chip, int cmd)
{
	int ret;
	struct denali_nand_prv *prv = chip->priv;

	INIT_COMPLETION(prv->cmd_complete);

	denali_index_data_write(chip, MAP10 << 12 | cmd << 8 | 0x1);

	denali_index_control_set(chip, MAP10 << 26 | prv->chipsel << 24 |
				(CPHYSADDR(buffers) & 0xffff0000) >> 8);
	denali_index_data_write(chip, MAP10 << 12 | 0x2 << 8);

	denali_index_control_set(chip, MAP10 << 26 | prv->chipsel << 24 |
				(CPHYSADDR(buffers) & 0x0000ffff) << 8);
	denali_index_data_write(chip, MAP10 << 12 | 0x3 << 8);

	denali_index_control_set(chip, MAP10 << 26 | prv->chipsel << 24 |
				0x1 << 16 | 0x40 << 8);
	denali_index_data_write(chip, MAP10 << 12 | 0x4 << 8);

	/* INTR Enable DNA CMD COMP */
	if (prv->chipsel == 0)
		denali_writel(chip, NANDF_INTR_EN0, 0x4);
	else
		denali_writel(chip, NANDF_INTR_EN1, 0x4);

	/* Wait DMA Complete */
	ret = wait_for_completion_interruptible_timeout(&prv->cmd_complete, HZ);

	/* DMA Disable */
	denali_writel(chip, NANDF_DMA_ENABLE, 0x0);

	/* INTR Disable */
	if (prv->chipsel == 0)
		denali_writel(chip, NANDF_INTR_EN0, 0x0);
	else
		denali_writel(chip, NANDF_INTR_EN1, 0x0);

	if (ret == 0) {
		printk(KERN_ERR "controller timed out\n");
		return ret;
	}

	return 0;
}

static void denali_nand_cmdfunc(struct mtd_info *mtd,
				    unsigned command,
				    int column,
				    int page_addr)
{
	struct nand_chip *chip = mtd->priv;
	struct denali_nand_prv *prv = chip->priv;

	switch (command) {
	case NAND_CMD_SEQIN:
	case NAND_CMD_PAGEPROG:
	case NAND_CMD_CACHEDPROG:
		break;

	case NAND_CMD_READOOB:
		/* DMA Enable */
		denali_writel(chip, NANDF_DMA_ENABLE, 0x1);

		/* Page Address Encoding */
		denali_index_control_set(chip, MAP10 << 26 |
				prv->chipsel << 24 | page_addr);
		break;

	case NAND_CMD_READ0:
		/* DMA Enable */
		denali_writel(chip, NANDF_DMA_ENABLE, 0x1);

		/* Page Address Encoding */
		denali_index_control_set(chip, MAP10 << 26 |
					prv->chipsel << 24 | page_addr);
		break;

	case NAND_CMD_ERASE1:
		/* Page Address Encoding */
		denali_index_control_set(chip, MAP10 << 26 |
					prv->chipsel << 24 | page_addr);
		/* Erase CMD */
		denali_index_data_write(chip, 0x1);
		wait_erase_complete(chip);
		break;

	case NAND_CMD_ERASE2:
		/* marge "NAND_CMD_ERASE1:" */
		break;

	case NAND_CMD_STATUS:
		/* status cmd cycle */
		denali_index_control_set(chip, MAP11 << 26 |
				prv->chipsel << 24 | COMMAND_CYCLE);
		denali_index_data_write(chip, NAND_CMD_STATUS);

		/* status cycle */
		denali_index_control_set(chip, MAP11 << 26 |
				prv->chipsel << 24 | STATUS_CYCLE);
		break;

	case NAND_CMD_READID:
		/* read id cmd cycle */
		denali_index_control_set(chip, MAP11 << 26 |
				prv->chipsel << 24 | COMMAND_CYCLE);
		denali_index_data_write(chip, NAND_CMD_READID);

		denali_index_control_set(chip, MAP11 << 26 |
				prv->chipsel << 24 | ADDRESS_CYCLE);
		denali_index_data_write(chip, 0x00);

		/* status cycle */
		denali_index_control_set(chip, MAP11 << 26 |
				prv->chipsel << 24 | STATUS_CYCLE);
		break;

	case NAND_CMD_RESET:
		/* reset cmd cycle */
		denali_index_control_set(chip, MAP11 << 26 |
				prv->chipsel << 24 | COMMAND_CYCLE);
		denali_index_data_write(chip, NAND_CMD_RESET);

		/* R/B pin of device check */
		denali_nand_dev_ready(mtd);
		break;
	}
}

/**
 * denali_nand_write_page - write one page
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @buf:	data buffer
 * @page:	page number to write
 * @cached:	cached programming
 * @raw:	use _raw version of write_page
 */
static int denali_nand_write_page(struct mtd_info *mtd, struct nand_chip *chip,
			   const uint8_t *buf, int page, int cached, int raw)
{
	struct denali_nand_prv *prv = chip->priv;

	/* DMA Enable */
	denali_writel(chip, NANDF_DMA_ENABLE, 0x1);

	/* Page Address Encoding */
	denali_index_control_set(chip, MAP10 << 26 |
				prv->chipsel << 24 | page);
	chip->ecc.write_page(mtd, chip, buf);

	return 0;
}

/**
 * nand_get_device - Get chip for selected access
 * @chip:	the nand chip descriptor
 * @mtd:	MTD device structure
 * @new_state:	the state which is requested
 *
 * Get the device and lock it for exclusive access
 */
static int
nand_get_device(struct nand_chip *chip, struct mtd_info *mtd, int new_state)
{
	spinlock_t *lock = &chip->controller->lock;
	wait_queue_head_t *wq = &chip->controller->wq;
	DECLARE_WAITQUEUE(wait, current);
 retry:
	spin_lock(lock);

	/* Hardware controller shared among independend devices */
	if (!chip->controller->active)
		chip->controller->active = chip;

	if (chip->controller->active == chip && chip->state == FL_READY) {
		chip->state = new_state;
		spin_unlock(lock);
		return 0;
	}
	if (new_state == FL_PM_SUSPENDED) {
		spin_unlock(lock);
		return (chip->state == FL_PM_SUSPENDED) ? 0 : -EAGAIN;
	}
	set_current_state(TASK_UNINTERRUPTIBLE);
	add_wait_queue(wq, &wait);
	spin_unlock(lock);
	schedule();
	remove_wait_queue(wq, &wait);
	goto retry;
}

/**
 * denali_nand_wait - wait until the command is done
 * @mtd:	MTD device structure
 * @chip:	NAND chip structure
 *
 * Wait for command done.
 */
static int denali_nand_wait(struct mtd_info *mtd, struct nand_chip *chip)
{
	denali_nand_dev_ready(mtd);

	return 0;
}

/**
 * nand_release_device - release chip
 * @mtd:	MTD device structure
 *
 * Deselect, release chip lock and wake up anyone waiting on the device
 */
static void nand_release_device(struct mtd_info *mtd)
{
	struct nand_chip *chip = mtd->priv;

	/* De-select the NAND device */
	chip->select_chip(mtd, -1);

	/* Release the controller and the chip */
	spin_lock(&chip->controller->lock);
	chip->controller->active = NULL;
	chip->state = FL_READY;
	wake_up(&chip->controller->wq);
	spin_unlock(&chip->controller->lock);
}

/**
 * denali_nand_block_bad - Read bad block marker from the chip
 * @mtd:	MTD device structure
 * @ofs:	offset from device start
 * @getchip:	0, if the chip is already selected
 *
 * Check, if the block is bad.
 */
static int denali_nand_block_bad(struct mtd_info *mtd, loff_t ofs, int getchip)
{
	int page, chipnr, res = 0;
	struct nand_chip *chip = mtd->priv;
	u16 bad;

	page = (int)(ofs >> chip->page_shift) & chip->pagemask;

	if (getchip) {
		chipnr = (int)(ofs >> chip->chip_shift);

		nand_get_device(chip, mtd, FL_READING);

		/* Select the NAND device */
		chip->select_chip(mtd, chipnr);
	}

	if (chip->options & NAND_BUSWIDTH_16) {
		chip->cmdfunc(mtd, NAND_CMD_READOOB, chip->badblockpos & 0xFE,
			      page);
		bad = cpu_to_le16(chip->read_word(mtd));
		if (chip->badblockpos & 0x1)
			bad >>= 8;
		if ((bad & 0xFF) != 0xff)
			res = 1;
	} else {
		chip->cmdfunc(mtd, NAND_CMD_READOOB, chip->badblockpos, page);
		if (buffers[mtd->writesize] != 0xff)
			res = 1;
	}

	if (getchip)
		nand_release_device(mtd);

	return res;
}

/*
 * denali_nand_read_page_raw - read raw page data without ecc
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @buf:	buffer to store read data
 */
static int denali_nand_read_page_raw(struct mtd_info *mtd,
				struct nand_chip *chip, uint8_t *buf)
{
	int ret, i;
	struct denali_nand_prv *prv = chip->priv;

	/* DMA Read */
	dma_cache_wback_inv((unsigned long)buffers,
				mtd->writesize + mtd->oobsize);
	ret = nand_dma_transfer(chip, READ);
	if (ret != 0)
		return ret;

	/* Sector data Denali Layout */
	for (i = 0; i < mtd->writesize / SECTOR_SIZE; i++)
		memcpy(buf + SECTOR_SIZE * i,
			buffers + (SECTOR_SIZE + prv->ecc) * i,
			SECTOR_SIZE);

	/* ECC Data Denali Layout */
	for (i = 0; i < 4 ; i++)
		memcpy(chip->oob_poi + prv->ecc * i, buffers + SECTOR_SIZE
			+ (SECTOR_SIZE + prv->ecc) * i, prv->ecc);

	/* Flags Data Denali Layout */
	memcpy(chip->oob_poi + mtd->oobsize - mtd->oobavail,
		buffers + mtd->writesize + mtd->oobsize - mtd->oobavail,
		mtd->oobavail);

	return 0;
}

/**
 * denali_nand_read_page_hwecc - hardware ecc based page read function
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @buf:	buffer to store read data
 */
static int denali_nand_read_page_hwecc(struct mtd_info *mtd,
				struct nand_chip *chip, uint8_t *buf)
{
	int ret, i;
	struct denali_nand_prv *prv = chip->priv;

	/* DMA Read */
	dma_cache_wback_inv((unsigned long)buffers,
				mtd->writesize + mtd->oobsize);
	ret = nand_dma_transfer(chip, READ);

	if (ret != 0)
		return ret;

	/* ECC correcte */
	if (denali_readl(chip, NANDF_ECCCORINFO_B01) &
		(0x1 << (7 + 8 * prv->chipsel))) {
			for (i = 0; i < mtd->writesize + mtd->oobsize; i++) {
				if (buffers[i] != 0xff) {
					mtd->ecc_stats.failed++;
					return -EIO;
				}
			}
	}

	/* Sector data Denali Layout */
	for (i = 0; i < mtd->writesize / SECTOR_SIZE; i++)
		memcpy(buf + SECTOR_SIZE * i,
			buffers + (SECTOR_SIZE + prv->ecc) * i,
			SECTOR_SIZE);

	/* ECC Data Denali Layout */
	for (i = 0; i < 4 ; i++)
		memcpy(chip->oob_poi + prv->ecc * i, buffers + SECTOR_SIZE
			+ (SECTOR_SIZE + prv->ecc) * i, prv->ecc);

	/* Flags data Denali Layout */
	memcpy(chip->oob_poi + mtd->oobsize - mtd->oobavail,
		buffers + mtd->writesize + mtd->oobsize - mtd->oobavail,
		mtd->oobavail);

	return 0;
}

/**
 * denali_nand_write_page_hwecc - hardware ecc based page write function
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @buf:	data buffer
 */
static void denali_nand_write_page_hwecc(struct mtd_info *mtd,
				struct nand_chip *chip, const uint8_t *buf)
{
	int i;
	struct denali_nand_prv *prv = chip->priv;

	/* Sector data Denali Layout */
	for (i = 0; i < mtd->writesize / SECTOR_SIZE; i++)
		memcpy(buffers + (SECTOR_SIZE + prv->ecc) * i,
			buf + SECTOR_SIZE * i, SECTOR_SIZE);

	/* Flags data Denali Layout */
	memcpy(buffers + mtd->writesize + mtd->oobsize - mtd->oobavail,
		chip->oob_poi + mtd->oobsize - mtd->oobavail, mtd->oobavail);

	/* DMA Write */
	dma_cache_wback_inv((unsigned long)buffers,
				mtd->writesize + mtd->oobsize);
	nand_dma_transfer(chip, WRITE);
}

/**
 * denali_nand_read_oob_std - the most common OOB data read function
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @page:	page number to read
 * @sndcmd:	flag whether to issue read command or not
 */
static int denali_nand_read_oob_std(struct mtd_info *mtd,
			     struct nand_chip *chip, int page, int sndcmd)
{
	struct denali_nand_prv *prv = chip->priv;
	int ret, i;

	if (sndcmd) {
		chip->cmdfunc(mtd, NAND_CMD_READOOB, 0, page);
		sndcmd = 0;
	}

	/* DMA Read */
	dma_cache_wback_inv((unsigned long)buffers,
				mtd->writesize + mtd->oobsize);
	ret = nand_dma_transfer(chip, READ);
	if (ret != 0)
		return ret;

	/* ECC Data Denali Layout */
	for (i = 0; i < 4; i++)
		memcpy(chip->oob_poi + prv->ecc * i, buffers + SECTOR_SIZE
			+ (SECTOR_SIZE + prv->ecc) * i, prv->ecc);

	/* Flags Data Denali Layout */
	memcpy(chip->oob_poi + mtd->oobsize - mtd->oobavail,
		buffers + mtd->writesize + mtd->oobsize - mtd->oobavail,
		mtd->oobavail);

	return sndcmd;
}

/**
 * denali_nand_write_oob_std - the most common OOB data write function
 * @mtd:	mtd info structure
 * @chip:	nand chip info structure
 * @page:	page number to write
 */
static int denali_nand_write_oob_std(struct mtd_info *mtd,
				struct nand_chip *chip, int page)
{
	int ret;
	const uint8_t *buf = chip->oob_poi;
	struct denali_nand_prv *prv = chip->priv;

	/* HW ECC Disable */
	denali_writel(chip, NANDF_ECC_EN, 0x0);

	/* DMA Enable */
	denali_writel(chip, NANDF_DMA_ENABLE, 0x1);

	/* Page Address Encoding */
	denali_index_control_set(chip, MAP10 << 26 |
			prv->chipsel << 24 | page);

	/* DMA Read */
	dma_cache_wback_inv((unsigned long)buffers,
				mtd->writesize + mtd->oobsize);
	ret = nand_dma_transfer(chip, READ);
	if (ret != 0)
		return ret;

	/* Flag data Denali Layout */
	memcpy(buffers + mtd->writesize + mtd->oobsize - mtd->oobavail,
		buf + mtd->oobsize - mtd->oobavail, mtd->oobavail);

	/* R/B pin of device check */
	denali_nand_dev_ready(mtd);


	/* DMA Enable */
	denali_writel(chip, NANDF_DMA_ENABLE, 0x1);

	/* Page Address Encoding */
	denali_index_control_set(chip, MAP10 << 26 |
			prv->chipsel << 24 | page);

	/* DMA Write */
	dma_cache_wback_inv((unsigned long)buffers,
				mtd->writesize + mtd->oobsize);
	ret = nand_dma_transfer(chip, WRITE);

	/* HW ECC Enable */
	denali_writel(chip, NANDF_ECC_EN, 0x1);

	return ret;
}

static irqreturn_t denali_nand_interrupt(int irq, void *dev_instance)
{
	struct mtd_info *mtd = (struct mtd_info *)dev_instance;
	struct nand_chip *chip = mtd->priv;
	struct denali_nand_prv *prv = chip->priv;

	/* Interrupt status register clear */
	if (prv->chipsel == 0) {
		if (denali_readl(chip, NANDF_INTR_STATUS0) & DMA_COMP) {
			denali_writel(chip, NANDF_INTR_STATUS0, 0x4);
			complete(&prv->cmd_complete);
		}
	} else {
		if (denali_readl(chip, NANDF_INTR_STATUS1) & DMA_COMP) {
			denali_writel(chip, NANDF_INTR_STATUS1, 0x4);
			complete(&prv->cmd_complete);
		}
	}

	return IRQ_HANDLED;
}

static void hwcontrol_init(struct nand_chip *chip)
{
	struct denali_nand_prv *prv = chip->priv;
	u32 tmp;

	/* DMA Disable */
	denali_writel(chip, NANDF_DMA_ENABLE, 0x0);

	/* ECC correction */
	tmp = readl(prv->b_regs + NANDF_BOOTSTRAP_INFO) >> 1;

	switch (tmp & 0x7) {
	/* 4bit ecc correction */
	case 0:
		prv->ecc = 8;
		break;
	/* 8bit ecc correction */
	case 1:
		prv->ecc = 14;
		break;
	/* 15bit ecc correction */
	case 2:
		prv->ecc = 26;
		break;
	default:
		printk(KERN_ERR ": not support ECC Correction!\n");
		prv->ecc = 0;
		break;
	}

	/* Non-ONFI compliant device */
	denali_writel(chip, NANDF_ONFI_DEVICE_NO_OF_LUNS, 0x0);

	/* Main + Spare mode */
	denali_writel(chip, NANDF_TRANSFER_SPARE, 0x1);

	/* Ready/Busy Pin is enable */
	tmp = denali_readl(chip, NANDF_RD_PIN_EN);
	denali_writel(chip, NANDF_RD_PIN_EN, tmp | 0x1 << prv->chipsel);

	/* Global Interrupt enable */
	denali_writel(chip, NANDF_GLOBAL_INT_EN, 0x1);

	/* HW ECC Enable */
	denali_writel(chip, NANDF_ECC_EN, 0x1);

	/* DMA interrupt clear */
	denali_writel(chip, NANDF_DMA_INTR, 0x0);

	/* Device reset */
	if (prv->chipsel)
		denali_writel(chip, NANDF_DEVICE_RESET, 0x1 << prv->chipsel);

	/* INTR0 Disable and clear */
	denali_writel(chip, NANDF_INTR_EN0, 0x0);
	denali_writel(chip, NANDF_INTR_STATUS0, 0xffff);

	/* INTR1 Disable and clear */
	denali_writel(chip, NANDF_INTR_EN1, 0x0);
	denali_writel(chip, NANDF_INTR_STATUS1, 0xffff);
}

static int __devinit denali_nand_probe(struct platform_device *pdev)
{
	struct mtd_info *mtd;
	struct nand_chip *chip;
	struct denali_nand_prv *dev;
	struct resource *mem;
	const char *part_type;
	struct denali_nand_platform_data *pdata = pdev->dev.platform_data;
	int irq, r;

#ifdef CONFIG_MTD_PARTITIONS
	int mtd_parts_nb = 0;
	struct mtd_partition *mtd_parts;
#endif

	printk(KERN_INFO "NAND Flash ROM driver for denali\n");

	/* private data  */
	dev = kmalloc(sizeof(struct denali_nand_prv), GFP_KERNEL);
	if (!dev) {
		dev_err(&pdev->dev, "Unable to allocate\n");
		return -ENOMEM;
	}
	memset(dev, 0, sizeof(struct denali_nand_prv));

	/* NOTE: driver uses the static register mapping */
	/* Get Resource Controller register */
	mem = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!mem) {
		dev_err(&pdev->dev, "no mem resource?\n");
		r = -EINVAL;
		goto err_free_mem;
	}

	dev->c_regs = ioremap(mem->start, resource_size(mem));
	if (dev->c_regs == NULL) {
		dev_err(&pdev->dev, "failure mapping io resources\n");
		r = -EBUSY;
		goto err_free_mem;
	}

	/* Get Resource Controller Indexed-Addressing register */
	mem = platform_get_resource(pdev, IORESOURCE_MEM, 1);
	if (!mem) {
		dev_err(&pdev->dev, "no mem resource?\n");
		r = -EINVAL;
		goto err_iounmap_c;
	}

	dev->i_regs = ioremap(mem->start, resource_size(mem));
	if (dev->i_regs == NULL) {
		dev_err(&pdev->dev, "no mem resource?\n");
		r = -EBUSY;
		goto err_iounmap_c;
	}

	/* Get Resource Local Bridge register */
	mem = platform_get_resource(pdev, IORESOURCE_MEM, 2);
	if (!mem) {
		dev_err(&pdev->dev, "no mem resource?\n");
		r = -EINVAL;
		goto err_iounmap_i;
	}

	dev->b_regs = ioremap(mem->start, resource_size(mem));
	if (dev->b_regs == NULL) {
		dev_err(&pdev->dev, "no mem resource?\n");
		r = -EBUSY;
		goto err_iounmap_i;
	}

	irq = platform_get_irq(pdev, 0);
	if (irq < 0) {
		dev_err(&pdev->dev, "no irq resource?\n");
		r = irq; /* -ENXIO */
		goto err_iounmap_b;
	}

	init_completion(&dev->cmd_complete);
	dev->irq = irq;
	dev->chipsel =  pdev->id;
	chip = &dev->chip;

	platform_set_drvdata(pdev, dev);
	chip->priv = dev;
	mtd = &dev->mtd;
	mtd->priv = chip;
	hwcontrol_init(chip);

	r = request_irq(dev->irq, &denali_nand_interrupt, IRQF_SHARED,
				pdev->name, &dev->mtd);
	if (r) {
		dev_err(&pdev->dev, "failure requesting irq %i\n", dev->irq);
		goto err_iounmap_b;
	}

	chip->cmdfunc = denali_nand_cmdfunc;
	chip->select_chip = denali_nand_select_chip;
	chip->read_byte = denali_nand_read_byte;
	chip->read_word = denali_nand_read_word;
	chip->write_buf = denali_nand_write_buf;
	chip->read_buf = denali_nand_read_buf;
	chip->cmd_ctrl = denali_nand_cmd_ctrl;
	chip->dev_ready = denali_nand_dev_ready;
	chip->waitfunc = denali_nand_wait;

	chip->write_page = denali_nand_write_page;
	chip->ecc.read_page_raw = denali_nand_read_page_raw;
	chip->ecc.write_page_raw = denali_nand_write_page_hwecc;

	chip->block_bad = denali_nand_block_bad;
	chip->ecc.mode = NAND_ECC_HW;
	chip->ecc.read_page = denali_nand_read_page_hwecc;
	chip->ecc.write_page = denali_nand_write_page_hwecc;
	chip->ecc.write_oob = denali_nand_write_oob_std;
	chip->ecc.read_oob = denali_nand_read_oob_std;

	chip->controller = &denali_hwcontrol;
	spin_lock_init(&chip->controller->lock);
	init_waitqueue_head(&chip->controller->wq);

	chip->options = NAND_USE_FLASH_BBT;

	if (pdata->page == LARGE_PAGE) {
		/* large page */
		chip->ecc.layout = &nand_oob_64;
		chip->bbt_td = &bbt_main_descr;
		chip->bbt_md = &bbt_mirror_descr;
	} else if (pdata->page == VERY_LARGE_PAGE) {
		/* very large page */
		chip->ecc.layout = &nand_oob_128;
		chip->bbt_td = &bbt_main_descr_vlp;
		chip->bbt_md = &bbt_mirror_descr_vlp;
	} else {
		/* not support page */
		r = -ENXIO;
		goto err_free_irq;
	}
	chip->ecc.bytes = dev->ecc;
	chip->ecc.size = SECTOR_SIZE;

	/* Scan to find existence of the device */
	if (nand_scan(mtd, 1)) {
		r = -ENXIO;
		goto err_free_irq;
	}

#ifdef CONFIG_MTD_PARTITIONS
	mtd->name =  "denali_nand";
#ifdef CONFIG_MTD_CMDLINE_PARTS
	mtd_parts_nb = parse_mtd_partitions(mtd,
					partition_probes,
					&mtd_parts,
					0);
#endif
	if (mtd_parts_nb > 0)
		part_type = "command line";
	else {
		mtd_parts	= pdata->parts;
		mtd_parts_nb	= pdata->nr_parts;
		part_type	= "static";
	}

	/* Register the partitions */
	printk(KERN_NOTICE "Using %s partition definition\n", part_type);

	/* Creating 1 MTD partitions on "emma-nand" */
	add_mtd_partitions(mtd, mtd_parts, mtd_parts_nb);
#endif
	add_mtd_device(mtd);

	return 0;

err_free_irq:
	free_irq(dev->irq, dev);
err_iounmap_b:
	iounmap(dev->b_regs);
err_iounmap_i:
	iounmap(dev->i_regs);
err_iounmap_c:
	iounmap(dev->c_regs);
err_free_mem:
	kfree(dev);

	return r;
}

static int __devexit denali_nand_remove(struct platform_device *pdev)
{
	struct denali_nand_prv *dev = platform_get_drvdata(pdev);
	struct mtd_info *mtd = &dev->mtd;

	nand_release(mtd);

	free_irq(dev->irq, dev);
	iounmap(dev->b_regs);
	iounmap(dev->i_regs);
	iounmap(dev->c_regs);
	kfree(dev);

	return 0;
}

static struct platform_driver denali_nand_driver = {
	.remove		= __devexit_p(denali_nand_remove),
	.driver		= {
		.name	= "denali_nand",
		.owner	= THIS_MODULE,
	},
};

static int __init denali_nand_init_driver(void)
{
	return platform_driver_probe(&denali_nand_driver, denali_nand_probe);
}

static void __exit denali_nand_exit_driver(void)
{
	platform_driver_unregister(&denali_nand_driver);
}

module_init(denali_nand_init_driver);
module_exit(denali_nand_exit_driver);

MODULE_AUTHOR("Renesas Electronics Corporation");
MODULE_DESCRIPTION("NAND flash driver for Denali Controller");
MODULE_LICENSE("GPL v2");

