/*
 * OHCI HCD (Host Controller Driver) for USB.
 *
 * Copyright (C) 2009  NEC Electronics Corporation
 *
 * Bus Glue for NEC EMMAXXX built-in OHCI host controller
 *
 * Based on fragments of previous drivers:
 * hcd-pci.c by David Brownell <dbrownell@users.sourceforge.net>
 * ohci-sa1111.c by Christopher Hoover <ch@hpl.hp.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * (version 2) as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <linux/platform_device.h>
#include <linux/pci_regs.h>
#include <linux/usb.h>

static int emma_pseudo_pci_init(struct platform_device *pdev)
{
	struct resource *res, *cfg;
	void __iomem *base;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);

	cfg = platform_get_resource(pdev, IORESOURCE_MEM, 1);
	if (!cfg)
		return -ENXIO;
	base = ioremap(cfg->start, resource_size(cfg));
	if (!base)
		return -ENOMEM;

	/* Base Address Register */
	writel(res->start, base + PCI_BASE_ADDRESS_0);

	/* Command */
	writel(PCI_COMMAND_MASTER | PCI_COMMAND_MEMORY, base + PCI_COMMAND);

	iounmap(base);
	return 0;
}

static int emma_start_hc(struct platform_device *pdev)
{
	pr_debug(__FILE__ ": starting EMMA OHCI USB Controller\n");

	return emma_pseudo_pci_init(pdev);
}

static void emma_stop_hc(struct platform_device *pdev)
{
	pr_debug(__FILE__ ": stopping EMMA OHCI USB Controller\n");
}

/*-------------------------------------------------------------------------*/

/* configure so an HC device and id are always provided */
/* always called with process context; sleeping is OK */

/**
 * usb_hcd_emma_probe - initialize EMMA built-in HCDs
 * Context: !in_interrupt()
 *
 * Allocates basic resources for this USB host controller, and
 * then invokes the start() method for the HCD associated with it
 * through the hotplug entry's driver_data.
 *
 * Store this function in the HCD's struct platform_driver as probe().
 */
static int usb_hcd_emma_probe(const struct hc_driver *driver,
			      struct platform_device *pdev)
{
	struct resource *res;
	struct usb_hcd	*hcd;
	int irq, ret;

	/*
	 * Basic sanity checks first.  Board-specific init logic should
	 * have initialized these three resources and probably board
	 * specific platform_data.  We don't probe for IRQs, and do only
	 * minimal sanity checking.
	 */
	if (pdev->num_resources != 3) {
		dev_err(&pdev->dev, "hcd probe: invalid num_resources: %i\n",
			pdev->num_resources);
		return -ENODEV;
	}

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		dev_err(&pdev->dev, "Found HC with no register addr!\n");
		return -ENODEV;
	}

	irq = platform_get_irq(pdev, 0);
	if (irq <= 0) {
		dev_err(&pdev->dev, "Found HC with no IRQ!\n");
		return -ENODEV;
	}

	hcd = usb_create_hcd(driver, &pdev->dev, "emma-ohci");
	if (!hcd)
		return -ENOMEM;
	hcd->rsrc_start = (u64)res->start;
	hcd->rsrc_len = (u64)resource_size(res);

	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len, hcd_name)) {
		pr_debug("request_mem_region failed\n");
		ret = -EBUSY;
		goto err1;
	}

	hcd->regs = ioremap(hcd->rsrc_start, hcd->rsrc_len);
	if (!hcd->regs) {
		dev_err(&pdev->dev, "ioremap failed\n");
		ret = -ENOMEM;
		goto err2;
	}

	ret = emma_start_hc(pdev);
	if (ret) {
		dev_err(&pdev->dev, "Failed to start HC\n");
		goto err3;
	}

	ohci_hcd_init(hcd_to_ohci(hcd));

	ret = usb_add_hcd(hcd, irq, IRQF_SHARED);
	if (ret)
		goto err4;

	return 0;

 err4:
	emma_stop_hc(pdev);
 err3:
	iounmap(hcd->regs);
 err2:
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
 err1:
	usb_put_hcd(hcd);
	return ret;
}

/* may be called without controller electrically present */
/* may be called with controller, bus, and devices active */

/**
 * usb_hcd_emma_remove - shutdown processing for EMMA built-in HCDs
 * @dev: USB Host Controller being removed
 * Context: !in_interrupt()
 *
 * Reverses the effect of usb_hcd_emma_probe(), first invoking
 * the HCD's stop() method.  It is always called from a thread
 * context, normally "rmmod", "apmd", or something similar.
 *
 * Store this function in the HCD's struct platform_driver as remove().
 */
static void usb_hcd_emma_remove(struct usb_hcd *hcd,
				struct platform_device *pdev)
{
	usb_remove_hcd(hcd);
	emma_stop_hc(pdev);
	iounmap(hcd->regs);
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
	usb_put_hcd(hcd);
}

/*-------------------------------------------------------------------------*/

static int __devinit ohci_emma_start(struct usb_hcd *hcd)
{
	struct ohci_hcd	*ohci = hcd_to_ohci(hcd);
	int		ret;

	ret = ohci_init(ohci);
	if (ret  < 0)
		return ret;

	ret = ohci_run(ohci);
	if (ret < 0) {
		err("can't start %s", hcd->self.bus_name);
		ohci_stop(hcd);
		return ret;
	}
	return 0;
}

static const struct hc_driver ohci_emma_hc_driver = {
	.description		= hcd_name,
	.product_desc		= "EMMA OHCI",
	.hcd_priv_size		= sizeof(struct ohci_hcd),

	/*
	 * generic hardware linkage
	 */
	.irq			= ohci_irq,
	.flags			= HCD_USB11 | HCD_MEMORY,

	/*
	 * basic lifecycle operations
	 */
	.start			= ohci_emma_start,
	.stop			= ohci_stop,
	.shutdown		= ohci_shutdown,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue		= ohci_urb_enqueue,
	.urb_dequeue		= ohci_urb_dequeue,
	.endpoint_disable	= ohci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number	= ohci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data	= ohci_hub_status_data,
	.hub_control		= ohci_hub_control,
#ifdef	CONFIG_PM
	.bus_suspend		= ohci_bus_suspend,
	.bus_resume		= ohci_bus_resume,
#endif
	.start_port_reset	= ohci_start_port_reset,
};

/*-------------------------------------------------------------------------*/

static int ohci_hcd_emma_drv_probe(struct platform_device *pdev)
{
	int ret;

	if (usb_disabled())
		return -ENODEV;

	ret = usb_hcd_emma_probe(&ohci_emma_hc_driver, pdev);
	return ret;
}

static int ohci_hcd_emma_drv_remove(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);

	usb_hcd_emma_remove(hcd, pdev);
	return 0;
}

/*-------------------------------------------------------------------------*/

static struct platform_driver ohci_hcd_emma_driver = {
	.probe		= ohci_hcd_emma_drv_probe,
	.remove		= ohci_hcd_emma_drv_remove,
	.shutdown	= usb_hcd_platform_shutdown,
	.driver = {
		.name	= "emma-ohci",
		.owner	= THIS_MODULE,
	},
};

MODULE_ALIAS("platform:emma-ohci");
